#!/usr/bin/env python3
import sys, os, re

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

import xmlNode

datePat = re.compile("\d\d\d\d-\d\d-\d\d(?:_\d+)?")
soughtKeys = ["shortPathToBenchmark", "shortPathToBenchmarkDir"]

def extract(pathToInfoFile, nodePath=""):
  if not os.path.isfile(pathToInfoFile):
    print("'%s' either does not exist or is not a file" % pathToInfoFile)
    sys.exit(1)
  # else
  mn = xmlNode.parseXml(pathToInfoFile)
  if not mn.findChild(nodePath):
    print("'%s' does not lead to an node in '%s'"  % (nodePath, pathToInfoFile))
    sys.exit(1)
  # else
  mn = mn.findChild(nodePath)

  foundKeys = {}  # dictionary that will hold all keys for which we found at least 1 match
  bms = {}        # dictionary that will hold all unique benchmark-dates
  for sn in mn:
    for line in sn.text:
      for soughtKey in soughtKeys:
        if line.startswith(soughtKey):
          foundKeys[soughtKey] = None
          key, val = line.split(":",1)
          dates = datePat.findall(val)
          if dates:
            bms[dates[0]] = None
          else:
            print("warning: text of '%s' contains key '%s' but this key contains no date pattern" % (sn.getPathBelowRoot(), soughtKey))

  print("keys found: %s" % ", ".join(list(foundKeys.keys())))

  bms = list(bms.keys())  # transform to list
  bms.sort()
  for bm in bms:
    print(bm)

def usage():
  print("This program examines a info-file and prints a list of distinct")
  print("dates (invocation dirnames) to which the info-file makes reference")
  print("in its 'shortPathToBenchmark' and 'shortPathToBenchmarkDir' fields.")
  print("")
  print("usage: $ ./extractUniqueBenchmarks.py [path-to-info-file] [node-path]")
  print("")
  print("where 'path-to-info-file' is a path to the '.info' file and 'node-path'")
  print("is an optional argument that specifies a node below the master node of")
  print("the info file. If this argument is provided, the extraction of benchmark")
  print("dates will begin at the specified node.")
  sys.exit(0)

if __name__ == "__main__":
  if len(sys.argv) != 2 and len(sys.argv) != 3:
    usage()
  # else
  if len(sys.argv) == 2:
    extract(sys.argv[1])
  else:  # it must be 3
    extract(sys.argv[1], sys.argv[2])



"""Python API for FlashXTest"""

import os
from .. import lib
from .. import backend


def run(**apiDict):
    """
    Run a list of tests from test.info in current working directory
    """
    # Cache the value to current directory and set it as
    # testDir in apiDict
    apiDict["testDir"] = os.getcwd()

    # Cache the value of user Config file and store it as
    # pathToConfig in apiDict
    apiDict["pathToConfig"] = apiDict["testDir"] + "/config"

    # Set path to Info
    apiDict["pathToInfo"] = apiDict["testDir"] + "/test.info"

    # Environment variable for OpenMP
    # Set the default value. Each test
    # can override this from xml file
    os.environ["OMP_NUM_THREADS"] = str(1)

    # Get mainDict
    mainDict = lib.config.getMainDict(apiDict)

    # Parse test.info and create a testList
    jobList = []
    lib.info.jobListFromNode(
        backend.FlashTest.lib.xmlNode.parseXml(apiDict["pathToInfo"]), jobList
    )

    # Build sfocu for performing checks with baseline data
    # for Composite and Comparison tests
    lib.run.buildSFOCU(mainDict)

    # Run flashTest - actually call the backend flashTest.py here
    lib.run.flashTest(mainDict, jobList)



"""Python API for FlashXTest"""

import os
from .. import lib
from .. import backend


def setup_suite(**apiDict):
    """
    Setup test.info from a list of suites

    Arguments
    ---------
    testSuite : Name of the testSuite file
    """
    # Cache the value to current directory and set it as
    # testDir in apiDict
    apiDict["testDir"] = os.getcwd()

    # Cache the value of user Config file and store it as
    # pathToConfig in apiDict
    apiDict["pathToConfig"] = apiDict["testDir"] + "/config"

    # Get mainDict for performing tests. This will read
    # the user Config file and set values that
    # were not provided in apiDict and override values
    # that were
    mainDict = lib.config.getMainDict(apiDict)

    # Get specList from suite files
    specList = lib.suite.parseSuite(mainDict)

    # Create a test.info file for flashTest backend
    lib.info.createInfo(mainDict, specList)



from . import flashTestParser
from . import archive
from . import getProcessResults
from . import layeredDict
from . import logfile
from . import testObject
from . import xmlNode
from . import pidFile
from . import remoteFile



#!/usr/bin/env python3
import sys, os

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

import flashTestParser as parser

def go(startDir):

  def _delete():
    items = os.listdir(".")
    cwd = os.getcwd()
    deletedFiles = []
    notFound     = []
    for item in items:
      if os.path.isdir(item):
        os.chdir(item)
        _delete()
        os.chdir(cwd)
      elif item == "files_to_delete":
        filesToDelete = parser.fileToList(item)
        os.remove("files_to_delete")

        for fileToDelete in filesToDelete:
          if os.path.isfile(fileToDelete):
            os.remove(fileToDelete)
            deletedFiles.append(fileToDelete)
          else:
            notFound.append(fileToDelete + " (not found or not a file) ")

        deletedFiles.sort()
        notFound.sort()

        deletedFiles.extend(notFound)
        if deletedFiles:
          open("deleted_files","w").write("\n".join(deletedFiles))

  cwd = os.getcwd()
  os.chdir(startDir)
  _delete()
  os.chdir(cwd)


def usage(scriptName):
  lines = ["USAGE: %s [pathToDir]" % scriptName,
           "",
           "\"%s\" recursively searches 'pathToDir' for files called \"files_to_delete\"," % scriptName,
           "where 'pathToDir' is a path to a directory generated by FlashTest and \"files_to_delete\"",
           "is a newline-delimited list of filenames. Any files whose names appear in this list",
           "are deleted, and their filenames appended to a new file called \"deleted_files\". The",
           "original \"files_to_delete\" file will itself be deleted in all cases.",
           "",
           "Exactly which names appear in \"files_to_delete\" is determined by the Test object that",
           "originally generated the directory. See \"flashTest/lib/Tests.py\" for more details",
           ""]
  print("\n".join(lines))


if __name__ == "__main__":
  if len(sys.argv) != 2:
    usage(sys.argv[0])
  elif not os.path.isdir(sys.argv[1]):
    print("There is no directory called \"%s\"" % sys.argv[1])
  else:
    pathToDir = os.path.abspath(sys.argv[1])
    go(pathToDir)



"""FlashXTest library to interface with backend.FlashTest"""

import os, subprocess
import itertools
import glob
import argparse
import shlex

from .. import backend
from .. import lib


class TestSpec:
    """
    Class TestSpec to handle test specifications
    """

    def __init__(self):
        """
        Constructor
        """
        for attr in [
            "setupName",
            "nodeName",
            "setupOptions",
            "numProcs",
            "parfiles",
            "restartParfiles",
            "transfers",
            "environment",
            "debug",
        ]:
            setattr(self, attr, None)

    def getXmlText(self):
        """
        get Xml text from test specifications
        """

        # Create an empty list
        xmlText = []

        # Deal with parfile paths
        if self.parfiles:
            if self.parfiles == "<defaultParfiles>":
                pass
            else:
                parFileList = self.parfiles.split(" ")
                parFileList = [
                    "<pathToSimulations>" + "/" + self.setupName + "/tests/" + parfile
                    for parfile in parFileList
                ]
                self.parfiles = " ".join(parFileList)

        else:
            raise ValueError(
                lib.colors.FAIL
                + f"'parfiles' not defined for test {self.nodeName!r} for setup {self.setupName!r}"
            )

        # Deal with debug flags
        if self.debug:
            self.setupOptions = self.setupOptions + " -debug"

        # Deal with restartParfiles path
        if self.restartParfiles:
            parFileList = self.restartParfiles.split(" ")
            parFileList = [
                "<pathToSimulations>" + "/" + self.setupName + "/tests/" + parfile
                for parfile in parFileList
            ]
            self.restartParfiles = " ".join(parFileList)

        # Deal with environment variables
        if self.environment:
            self.environment = " ".join(
                list(itertools.chain.from_iterable(self.environment))
            )

        # append to xmlText
        for xmlKey in list(self.__dict__.keys()):
            if getattr(self, xmlKey):
                xmlText.append(f"{xmlKey}: {getattr(self, xmlKey)}")

        return xmlText


def __continuationLines(fin):
    for line in fin:
        line = line.rstrip("\n")
        while line.endswith("\\"):
            line = line[:-1] + next(fin).rstrip("\n")
        yield line


def parseSuite(mainDict):
    """
    Arguments
    ---------
    mainDict : Dicitionary for the API

    Returns
    specList : List of test specifications
    """
    # Set an empty dictionary to populate
    specList = []

    # Check if pathToSuites is defined, if not use
    # all *.suite files from the working directory
    if not mainDict["pathToSuites"]:
        mainDict["pathToSuites"] = glob.glob("*.suite")

    # Create a test suite parser
    suiteParser = argparse.ArgumentParser(description="Parser for test suite")
    suiteParser.add_argument("-t", "--test", help="Test node", type=str)
    suiteParser.add_argument("-np", "--nprocs", help="Num procs", type=int)
    suiteParser.add_argument(
        "-e", "--env", action="append", nargs="+", help="Environment variable", type=str
    )
    suiteParser.add_argument("--debug", action="store_true")
    suiteParser.set_defaults(debug=False, nprocs=1, test="", env=None)

    # Loop over all suite files and populate
    # suite dictionary
    for suiteFile in mainDict["pathToSuites"]:

        # Handle exceptions
        if not suiteFile.endswith(".suite"):
            raise ValueError(
                lib.colors.FAIL
                + f'[FlashXTest] File {suiteFile} must have a ".suite" suffix'
            )

        if not os.path.exists(suiteFile):
            raise ValueError(lib.colors.FAIL + f"[FlashXTest] Cannot find {suiteFile}")

        suiteList = []

        with open(suiteFile, "r") as sfile:
            for line in __continuationLines(sfile):
                suiteList.append(line.split("#")[0])

        suiteList = [spec for spec in suiteList if spec]

        for spec in suiteList:

            testSpec = TestSpec()
            testSpec.setupName = shlex.split(spec)[0]

            testArgs = suiteParser.parse_args(shlex.split(spec)[1:])
            testSpec.nodeName = testArgs.test

            for currSpec in specList:
                if testSpec.nodeName in currSpec.nodeName:
                    raise ValueError(
                        lib.colors.FAIL
                        + f"[FlashXTest] Duplicate for {testSpec.nodeName!r} detected in suite files"
                    )

            testSpec.numProcs = testArgs.nprocs
            testSpec.environment = testArgs.env
            testSpec.debug = testArgs.debug

            specList.append(testSpec)

    return specList



#!/usr/bin/env python3
import sys, os, subprocess, pwd
import re, socket
from time import asctime, strftime, localtime
from shutil import copy
import locale

def main():

    # some variables used below and what they represent:
    #
    #             version: the version number for this "FlashTest"
    #     pathToFlashTest: absolute path up to and including the directory
    #                      directly containing the flashTest.py script
    #        pathToOutdir: absolute path up to and including the directory
    #                      to which all FlashTest output will be directed
    #       pathToSiteDir: absolute path up to and including the directory
    #                      which contains all output for a given platform.
    #             siteDir: the name of the above directory; the basename of
    #                      'pathToSiteDir' and the fully qualified domain
    #                      name of the host computer
    # pathToInvocationDir: absolute path up to and including the directory
    #                      which contains all FlashTest output for a given
    #                      invocation.
    #       invocationDir: the name of the above directory; the basename of
    #                      'pathToInvocationDir'
    #          pathToInfo: absolute path up to and including the file
    #                      containing all "test.info" data for this platform
    #                 log: the Logfile instance for this invocation of
    #                      "FlashTest"
    #       pathToPidFile: absolute path to a file that stores the process id for this "FlashTest"

    version = "1.0"
    owner = pwd.getpwuid(os.getuid())[0]
    pathToFlashTest = os.path.dirname(os.path.abspath(sys.argv[0]))

    pathToTmpDir = os.getenv("TMPDIR", "/tmp")
    if pathToTmpDir == "":
        pathToTmpDir = "/tmp"
    pathToPidFile = os.path.join(pathToTmpDir, "flashTest.pid")

    def __abort0(msg):
        """
    called for errors very early in the initial invocation of FlashTest

    Because no logfile is instantiated yet, we try to record 'msg'
    in a file "ERROR" in the top-level FlashTest directory.

    This function does not attempt to remove a pidFile if it exists.
    """
        print(msg)
        # ignore failure of writing to the ERROR file in pathToFlashTest
        try:
            open(os.path.join(pathToFlashTest, "ERROR"), "a").write(
                "%s: %s\n" % (asctime(), msg)
            )
        except OSError:
            pass
        sys.exit(1)

    def __abort(msg):
        """
    called for errors in the initial invocation of FlashTest

    Because logfile may not be instantiated yet, we try to record 'msg'
    in a file "ERROR" in the top-level FlashTest directory.

    This function attempts to remove an existing pidFile on the way out.
    """
        print(msg)
        # ignore failure of writing to the ERROR file in pathToFlashTest
        try:
            open(os.path.join(pathToFlashTest, "ERROR"), "a").write(
                "%s: %s\n" % (asctime(), msg)
            )
        except OSError:
            pass
        finally:
            pidFile.unregisterPid(pathToPidFile)  # remove /tmp/flashTest.pid file before aborting
        sys.exit(1)

    def __deleteFiles(startDir):
        """
    Recursively search each directory under 'startDir' for text
    files named "files_to_delete", which will have been written
    by the test objects during run time, and which name files in
    that directory which should be deleted after the invocation
    has been archived.

    If a "files_to_delete" file is present, another file called
    "deleted_files" will be created, listing all files that were
    successfully deleted as well as any that were not found.

    The original "files_to_delete" file will itself be deleted
    in all cases.
    """

        def __delete():
            items = os.listdir(".")
            cwd = os.getcwd()
            deletedFiles = []
            notFound = []
            for item in items:
                if os.path.isdir(item):
                    os.chdir(item)
                    __delete()
                    os.chdir(cwd)
                elif item == "files_to_delete":
                    filesToDelete = parser.fileToList(item)
                    print("files to delete")
                    os.remove("files_to_delete")

                    for fileToDelete in filesToDelete:
                        if os.path.isfile(fileToDelete):
                            os.remove(fileToDelete)
                            deletedFiles.append(fileToDelete)
                        else:
                            notFound.append(
                                fileToDelete + " (not found or not a file) "
                            )

                    deletedFiles.extend(notFound)
                    if deletedFiles:
                        open("deleted_files", "w").write("\n".join(deletedFiles))

        cwd = os.getcwd()
        os.chdir(startDir)
        __delete()
        os.chdir(cwd)

    def __incErrors(whichStage):
        """
    Take a path and a stage 0-3, which corresponds to the stage
    of testing where the error occurred (0=setup, 1=compilation,
    2=execution, 3=testing) and note the error in "errors" files
    at all appropriate levels.
    """
        # masterDict["changedFromPrevious"] should be set by the test
        # object's tester component if we reached the testing stage and
        # found that the results of the test differed from those of the
        # previous invocation.
        # masterDict["sameAsPrevious"] should be set by the test
        # object's tester component if we reached the testing stage and
        # found that the results of the test differed from those of the
        # previous invocation.
        changedFromPrevious = masterDict.has_key("changedFromPrevious")
        if (whichStage == 3) and not changedFromPrevious:
            sameAsPrevious = masterDict.has_key("sameAsPrevious")
        else:
            sameAsPrevious = False

        # Note the error at the RUN LEVEL (if we've gotten that far)
        if masterDict.activeLayer >= 3:  # A run dir has been generated
            # with an "errors" file inside.
            # This "errors" file has only 2 or 3 lines:
            #
            # 0/1: no error/error during execution stage
            # 0/1: no error/error during testing stage
            # !/=: result of test differs from that of previous invocation/
            #                     is same as previous invocation
            #      3rd line may be absent, in which case comparison against
            #      previous result failed for reasons other than differences
            #      in the checkpoints; most likely because the "previous"
            #      checkpoint could not be found.
            errorsFile = os.path.join(pathToRunDir, "errors")
            if whichStage == 2:
                errorLines = ["1", "0"]
            elif whichStage == 3:
                errorLines = ["0", "1"]
                if changedFromPrevious:
                    errorLines.append("!")
                elif sameAsPrevious:
                    errorLines.append("=")

            open(errorsFile, "w").write("\n".join(errorLines))

        # Note the error at the BUILD LEVEL (if we've gotten that far)
        if masterDict.activeLayer >= 1:  # A build directory
            # hopefully has been generated
            # with an "errors" file inside.
            # This "errors" file has 6 or 7 lines:
            #
            # 0/1: no error/error during setup stage
            # 0/1: no error/error during compilation stage
            # 0/n: no errors/n errors during execution stage
            # 0/n: no errors/n errors during testing stage
            #  n : total number of runs attempted ( = num parfiles)
            # 0/n: no errors/n errors during testing that are same as before
            #      (included in numbers on 4th line)
            #  ! : results of 1 or more tests in this build differed
            #      from those of the previous invocation.
            errorsFile = os.path.join(pathToBuildDir, "errors")
            errorLines = parser.fileToList(errorsFile)

            if len(errorLines) == 6:
                if changedFromPrevious:
                    errorLines.append("!")
            else:  # can assume len(errorLines) is 7 with a "!" at the end
                # if this build has had *any* run whose results differed from
                # those of the previous invocation, set 'changedFromPrevious'
                # to True regardless of whether *this* run's results differed
                # or not. This will make sure the "!" gets put on the end of
                # the appropriate line at the invocation level (see below)
                changedFromPrevious = True

            # increment the value in stage 'whichStage' by 1
            errorLines[whichStage] = str(int(errorLines[whichStage]) + 1)
            if (whichStage == 3) and sameAsPrevious:
                errorLines[5] = str(int(errorLines[5]) + 1)
            open(errorsFile, "w").write("\n".join(errorLines))

            # we actually need these a few lines down
            # for the increment at the invocation level
            numExecErrs = int(errorLines[2])
            numTestErrs = int(errorLines[3])
            totalRuns = int(errorLines[4])
            numTestSameErrs = int(errorLines[5])
        #     else:   # if masterDict.activeLayer < 1
        #       numExecErrs = int(errorLines[2])
        #       numTestErrs = int(errorLines[3])
        #       totalRuns   = int(errorLines[4])
        #       numTestSameErrs = int(errorLines[5])

        # note the error at the INVOCATION LEVEL
        # This "errors" file contains names of all builds which yielded
        # some kind of error, noting the number of failed runs or tests
        # if the test failed in the execution or testing stages.
        errorsFile = os.path.join(pathToInvocationDir, "errors")
        errorLines = parser.fileToList(errorsFile)

        # compose err message
        if whichStage == 0:
            errMsg = "failed in setup"
        elif whichStage == 1:
            errMsg = "failed in compilation"
        elif whichStage == 2:
            errMsg = "%s/%s failed in execution" % (numExecErrs, totalRuns)
            if numTestErrs > 0:
                errMsg = "%s; %s/%s failed in testing" % (
                    errMsg,
                    numTestErrs,
                    totalRuns,
                )
                if numTestSameErrs > 0:
                    if numTestSameErrs == numTestErrs:
                        errMsg = errMsg + " as before"
                    else:
                        errMsg = "%s, % as before" % (errMsg, numTestSameErrs)
        elif whichStage == 3:
            errMsg = "%s/%s failed in testing" % (numTestErrs, totalRuns)
            if numTestSameErrs > 0:
                if numTestSameErrs == numTestErrs:
                    errMsg = errMsg + " as before"
                else:
                    errMsg = "%s, %s as before" % (errMsg, numTestSameErrs)
            if numExecErrs > 0:
                errMsg = "%s/%s failed in execution; %s" % (
                    numExecErrs,
                    totalRuns,
                    errMsg,
                )
        errMsg = "%s - %s" % (buildDir, errMsg)
        if changedFromPrevious:
            # add a "!" if changed from previous invocation's results
            errMsg = errMsg + " - !"

        for i in range(len(errorLines)):
            # if the invocation-level errors file already has
            # data about this build, replace it with 'errMsg'
            if errorLines[i].split(" - ", 1)[0] == buildDir:
                errorLines[i] = errMsg
                break
        else:
            # otherwise append it
            errorLines.append(errMsg)
            errorLines.sort()

        open(errorsFile, "w").write("\n".join(errorLines))

    ##########################
    ##  PARSE COMMAND LINE  ##
    ##########################

    # options to FlashTest which take no arguments
    # DEV should be able to adjust this in Config
    # especially now that "-u" is a custom option
    standAloneOpts = ["-t", "-u", "-v", "-vv", "-L", "--force"]

    # Retrieve a two-tuple for this invocation where the first element is
    # a dictionary of options (with arguments, if applicable) to FlashTest,
    # and the second element is a list of two-tuples, each containing a
    # test-path and a dictionary of options specific to that test.
    #
    # Graphically, the structure of the return value of 'parseCommandLine'
    # looks like:
    # ( { FlashTest-opt1: FlashTest-val1,
    #     FlashTest-opt2: FlashTest-val2 },
    #   [ ( test-path1, { test-path1-opt1: test-path1-val1,
    #                     test-path1-opt2: test-path1-val2 } ),
    #     ( test-path2, { test-path2-opt1: test-path2-val1 } ) ] )
    flashTestOpts, pathsAndOpts = parser.parseCommandLine(sys.argv[1:], standAloneOpts)

    # -------------------------------------------------------#
    # Check no other flashTest process is currently running
    # and/or prevent other instances from running
    # -------------------------------------------------------#
    if not pidFile.registerPid(pathToPidFile, "--force" in flashTestOpts):
        __abort0("It seems a flashTest process is already running. \n Use --force or delete %s to run this flashTest anyway." % (pathToPidFile,))

    # see if user has specified a separate "config"
    if "-c" in flashTestOpts:
        pathToConfig = flashTestOpts["-c"]
    else:
        pathToConfig = os.path.join(pathToFlashTest, "config")

    if "-e" in flashTestOpts:
        pathToExeScript = flashTestOpts["-e"]
    else:
        pathToExeScript = os.path.join(pathToFlashTest, "exeScript")

    #########################
    ##  PARSE CONFIG FILE  ##
    #########################

    if os.path.isfile(pathToConfig):
        try:
            configDict = parser.parseFile(pathToConfig)
        except Exception as e:
            __abort('Error opening "config" file\n' + str(e))
    else:
        __abort(
            'Configuration file "%s" does not exist or is not a file.' % pathToConfig
        )
    #########################
    ##  MASTER DICTIONARY  ##
    #########################

    # The master dictionary which we'll pass to the test object begins
    # life as a LayeredDict with initial keys and values copied from
    # 'configDict' (the copy is implicit in the initialization)
    masterDict = LayeredDict(configDict)

    masterDict["version"] = version
    masterDict["owner"] = owner
    masterDict["pathToFlashTest"] = pathToFlashTest
    masterDict["flashTestOpts"] = flashTestOpts
    masterDict["pathToExeScript"] = pathToExeScript

    ###########################
    ##  CHECK FOR -f OPTION  ##
    ###########################

    if "-f" in flashTestOpts:
        pathToJobFile = flashTestOpts["-f"]
        if not pathToJobFile:
            __abort(
                "The '-f' option must be followed by a path to a text file\n"
                + "consisting of a newline-delimited list of tests"
            )
        # else
        if not os.path.isfile(pathToJobFile):
            __abort('"%s" does not exist or is not a file.' % pathToJobFile)

        # Extend 'pathsAndOpts' to include test-paths read in from a file.
        # See comments to "parseCommandLine()" above for more details
        lines = parser.fileToList(pathToJobFile)

        # briefly re-join the 'lines' list into a single string,
        # with a single space inserted between list elements, then
        # re-split the new string by breaking on all whitespace.
        if lines:
            lines = re.split("\s*", " ".join(lines))

        pathsAndOpts.extend(parser.getPathsAndOpts(lines))

    if len(pathsAndOpts) == 0:
        __abort(
            "There are no tests to run. You must provide the name of at\n"
            + 'least one test or path to at least one "test.info" file.'
        )

    # else
    masterDict["pathsAndOpts"] = pathsAndOpts

    # get "test.info" data, if necessary
    masterInfoNode = None

    # Do a preliminary run through all our tests to see if any will require
    # data from a "test.info" file. If so, parse that data into an instance
    # of class XmlNode for use in all tests that require it, copying the
    # "test.info" file to the local machine from a remote host if necessary.
    if "-i" in flashTestOpts:
        # get 'pathToInfo' value from command-lines
        pathToInfo = flashTestOpts["-i"]
        if not os.path.isfile(pathToInfo):
            __abort('"%s" does not exist or is not a file.' % pathToInfo)
    else:
        # get 'pathToInfo' value from "config" or, if it's not there,
        # from the default location under 'pathToFlashTest'
        pathToInfo = masterDict.get(
            "pathToInfo", os.path.join(pathToFlashTest, "test.info")
        )

    # split 'pathToInfo' into its hostname and path segments
    if pathToInfo.count(":") > 0:
        infoHost, pathToInfo = pathToInfo.split(":", 1)
        if infoHost == "localhost":
            infoHost = ""
    else:
        infoHost = ""

    masterDict["infohost"] = infoHost
    masterDict["pathToInfo"] = pathToInfo

    # initializing variable
    masterDict["environment"] = ""

    for testPath, overrideOptions in pathsAndOpts:
        if testPath.count(os.sep) > 0:
            # At least one test in this invocation requires "test.info" data
            # so obtain the text of the "test.info" file whether it resides
            # on a remote computer or locally.
            if infoHost:
                # "test.info" file is on a remote computer
                pathToTmp = os.path.join(pathToFlashTest, "tmp")

                cmd = "scp %s:%s %s" % (infoHost, pathToInfo, pathToTmp)
                out, err, duration, exitStatus = getProcessResults(cmd)
                if err:
                    __abort(
                        'Unable to retrieve "%s:%s"\n%s' % (infoHost, pathToInfo, err)
                    )
                # else
                if exitStatus != 0:
                    __abort(
                        'Exit status %s indicates error retrieving "%s:%s"'
                        % (exitStatus, infoHost, pathToInfo)
                    )
                # else the file should have been transfered
                filename = os.path.basename(pathToInfo)
                if not os.path.isfile(os.path.join(pathToTmp, filename)):
                    __abort('File "%s" not found in "%s"' % (filename, pathToTmp))
                # else
                try:
                    masterInfoNode = parseXml(os.path.join(pathToTmp, filename))
                except Exception as e:
                    __abort('Error parsing "%s:%s"\n%s' % (infoHost, pathToInfo, e))

                # remove the copied file from "tmp"
                os.remove(os.path.join(pathToTmp, filename))

            else:
                # "test.info" file is local
                if not os.path.isfile(pathToInfo):
                    __abort('"%s" does not exist or is not a file' % pathToInfo)

                # else
                try:
                    masterInfoNode = parseXml(pathToInfo)
                except Exception as e:
                    __abort('Error parsing "%s"\n%s' % (pathToInfo, e))

            # 'masterInfoNode' now encapsulates all "test.info" data.
            # Break out of the for-loop.
            break

    #############################
    ##  GENERATE pathToOutdir  ##
    #############################

    pathToOutdir = flashTestOpts.get("-o", "")
    if not pathToOutdir:
        # check if we got a path to the outdir from the config file,
        # else use directory named "output" inside FlashTest directory
        pathToOutdir = masterDict.get(
            "pathToOutdir", os.path.join(pathToFlashTest, "output")
        )

    # create it if it doesn't exist
    if not os.path.exists(pathToOutdir):
        try:
            os.mkdir(pathToOutdir)
        except:
            __abort("Unable to create directory %s" % pathToOutdir)

    # make sure we have an absolute path
    pathToOutdir = os.path.abspath(pathToOutdir)

    masterDict["pathToOutdir"] = pathToOutdir

    ########################
    ##  GENERATE siteDir  ##
    ########################

    # try to determine the fully qualified domain name of the host
    if masterDict.has_key("FQHostname"):
        FQHostname = masterDict["FQHostname"]
    else:
        # DEV must allow a mechanism for dealing with aliases
        # DEV need a better mechanism for figuring out computer name
        # DEV when on a wireless network
        # DEV Also, if gethostbyaddr doesn't work, flashTest.py will
        # crash even tho it's not necessary - FQHostname only needs
        # to be defined if 'site' is not (see below)
        FQHostname = socket.gethostbyaddr(socket.gethostname())[0]
        masterDict["FQHostname"] = FQHostname
    # Determine value of "site". This will be the first element of
    # 'FQHostname' unless the user has specified it with the "-s"
    # command-line option or by declaring "site" in "config"
    if "-s" in flashTestOpts:
        # get the site from the command line
        siteDir = flashTestOpts["-s"]
    elif "site" in masterDict:
        # get the site from the config file
        siteDir = masterDict["site"]
    else:
        # 'siteDir' will be the first element of the fully-qualified
        # hostname as returned by socket.gethostname(). i.e. if the
        # fully-qualified hostname has a "." in it, 'siteDir' will
        # be everything that comes before that "."
        siteDir = FQHostname.split(".", 1)[0]

    pathToSiteDir = os.path.join(pathToOutdir, siteDir)

    # create it if it doesn't exist
    if not os.path.exists(pathToSiteDir):
        try:
            os.mkdir(pathToSiteDir)
        except:
            __abort("Unable to create directory %s" % pathToSiteDir)

    masterDict["pathToSiteDir"] = pathToSiteDir
    masterDict["siteDir"] = siteDir

    # TODO: Find an elegant way to do this
    # This is a hacky way to change 'flashSite' dir
    # through command line options and make it conistent
    # with 'siteDir'. 
    masterDict["flashSite"] = str(siteDir)

    ##############################
    ##  GENERATE invocationDir  ##
    ##############################

    # Generate a directory to hold all output from this invocation of FlashTest
    # Its name will be the current day's date, plus a suffix if necessary

    # start by getting the date in YYYY-MM-DD format
    dateStr = strftime("%Y-%m-%d", localtime())

    # get a list of any dirs in "output" that already start with dateStr
    dirs = [
        item
        for item in os.listdir(pathToSiteDir)
        if (
            os.path.isdir(os.path.join(pathToSiteDir, item))
            and item.startswith(dateStr)
        )
    ]

    if len(dirs) > 0:
        suffixes = [dir[len(dateStr) :].strip("_") for dir in dirs]
        suffixes.sort()
        highestSuffix = suffixes[len(suffixes) - 1]
        if highestSuffix == "":
            newSuffix = 2
        else:
            newSuffix = int(highestSuffix) + 1
        # there has already been at least one invocation
        # of FlashTest on this date, so add a suffix
        dateStr = "%s_%s" % (dateStr, newSuffix)
    masterDict["dateStr"] = dateStr
    invocationDir = dateStr
    pathToInvocationDir = os.path.join(pathToSiteDir, invocationDir)

    # create invocationDir
    try:
        os.mkdir(pathToInvocationDir)
    except:
        __abort("Unable to create directory %s" % pathToInvocationDir)

    # create a ".lock" file so that FlashTestView
    # will ignore it until the invocation is complete
    open(os.path.join(pathToInvocationDir, ".lock"), "w").write("")

    # Create a file "errors" which will record the details
    # of all errors that happen during this invocation.
    #
    # This file will be the basis of the floating statistics
    # box that appears in FlashTestView when the user hovers
    # over an invocation link (i.e. a date).
    #
    # See the comments to "__incErrors()" for more details
    open(os.path.join(pathToInvocationDir, "errors"), "w").write("")

    # Write a copy of the level-0 dictionary of 'masterDict'
    # (which at this point contains everything in 'masterDict')
    # into the invocation directory as a newline-delimited list
    # of key-value pairs
    keysAndVals = [
        "%s: %s" % (key, masterDict.dicts[0][key]) for key in masterDict.dicts[0]
    ]
    open(os.path.join(pathToInvocationDir, "masterDict"), "w").write(
        "\n".join(keysAndVals)
    )

    masterDict["pathToInvocationDir"] = pathToInvocationDir
    masterDict["invocationDir"] = invocationDir

    ###########################
    ##  INSTANTIATE LOGFILE  ##
    ###########################

    log = Logfile(pathToInvocationDir, "flash_test.log", "-v" in flashTestOpts)
    log.info("FlashTest v%s started by %s on %s\n" % (version, owner, asctime()), False)
    log.info(
        "Original command-line: %s %s" % (sys.argv[0], " ".join(sys.argv[1:])), False
    )
    log.info("This invocation: %s" % invocationDir, False)
    # find python version
    log.info("Python version info:", False)
    log.info(sys.version, True)
    try:
        cwd = os.getcwd()
    except Exception as e:
        __abort("Initial getcwd: " + str(e))
    try:  # find svn version
        os.chdir(pathToFlashTest)
        log.info("FlashTest version info:", False)
        p = subprocess.Popen(
            ["svn", "info", os.getcwd()], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        log.info(p.communicate()[0], True)
    except:
        pass
    os.chdir(cwd)
    log.brk()

    masterDict["log"] = log

    #######################################################
    ##  CHECK FOR PREVIOUS INVOCATION'S ARCHIVE RESULTS  ##
    #######################################################

    if "-t" not in flashTestOpts:
        previousArchiveLog = os.path.join(pathToSiteDir, "archive.log")
        if os.path.isfile(previousArchiveLog):
            archiveText = open(previousArchiveLog, "r").read()
            log.info(archiveText, False)
            log.brk()

    ######################
    ##  ARCHIVE OBJECT  ##
    ######################
    # provides methods for communication with the remote archive
    pathToLocalArchive = masterDict.get(
        "pathToLocalArchive", os.path.join(pathToFlashTest, "localArchive")
    )
    if not os.path.isabs(pathToLocalArchive):
        pathToLocalArchive = os.path.join(pathToFlashTest, pathToLocalArchive)
        log.note('Using "%s" as path to local archive.' % pathToLocalArchive)

    masterDict["pathToLocalArchive"] = pathToLocalArchive

    mainArchiveHost = ""
    pathToMainArchive = masterDict.get("pathToMainArchive", "")
    if pathToMainArchive:
        # work out if the main archive resides on a remote computer
        if pathToMainArchive.count(":") > 0:
            mainArchiveHost, pathToMainArchive = pathToMainArchive.split(":", 1)
            if mainArchiveHost == "localhost":
                mainArchiveHost = ""

    masterDict["mainArchiveHost"] = mainArchiveHost
    masterDict["pathToMainArchive"] = pathToMainArchive

    viewArchiveHost = ""
    pathToViewArchive = masterDict.get("pathToViewArchive", "")
    if pathToViewArchive:
        # work out if the view archive resides on a remote computer
        if pathToViewArchive.count(":") > 0:
            viewArchiveHost, pathToViewArchive = pathToViewArchive.split(":", 1)
            if viewArchiveHost == "localhost":
                viewArchiveHost = ""

    masterDict["viewArchiveHost"] = viewArchiveHost
    masterDict["pathToViewArchive"] = pathToViewArchive

    logArchiveHost = ""
    pathToLogArchive = masterDict.get("pathToLogArchive", "")
    if pathToLogArchive:
        # work out if the log archive resides on a remote computer
        if pathToLogArchive.count(":") > 0:
            logArchiveHost, pathToLogArchive = pathToLogArchive.split(":", 1)
            if logArchiveHost == "localhost":
                logArchiveHost = ""

    masterDict["logArchiveHost"] = logArchiveHost
    masterDict["pathToLogArchive"] = pathToLogArchive

    arch = Archive(log, masterDict)
    masterDict["arch"] = arch

    #######################
    ##  BEGIN FLASHTEST  ##
    #######################

    try:
        testObject = TestObject(masterDict)
    except Exception as e:
        __abort(str(e))

    try:
        buildDir = "ALL"
    except:
        pass

    # else testObject is now successfully instantiated
    if testObject.entryPoint1() == False:  # user's 1st chance to interrupt
        __incErrors(0)
    else:
        # outer loop over all test tuples, where each test tuple contains a
        # test name and a list of secondary "build tuples". Each build tuple,
        # in turn, consists of a short path to the relevant "test.info" file
        # (or the empty string if no info file was indicated), and the list of
        # override-options to the test, if any
        for testPath, overrideOptions in pathsAndOpts:
            # some variables used below and what they represent:
            # (see also comments at top of file)
            #
            #        testPath: xml-path to a node in "test.info" which contains
            #                  data relevant to a single FlashTest build
            #  pathToBuildDir: absolute path up to and including the directory
            #                  which contains all output for a single build, and
            #                  which contains one or more instances of 'runDir'
            #                  (see below).
            #        buildDir: The basename of 'pathToBuildDir', an immedidate
            #                  sub-directory of 'invocationDir'
            #    pathToRunDir: absolute path up to and including the directory
            #                  which directly contains all output for one run
            #                  of the executable against a single set of runtime
            #                  parameters.
            #          runDir: The basename of 'pathToRunDir', an immediate sub-
            #                  directory of 'buildDir'
            masterDict.setActiveLayer(1)
            masterDict["testPath"] = testPath
            masterDict["overrideOptions"] = overrideOptions

            # start log entry for this build
            log.brk()
            shortPathToNode = os.path.normpath(os.path.join(siteDir, testPath))

            if testPath.count(os.sep) > 0:
                thisNode = masterInfoNode.findChild(shortPathToNode)

                if not thisNode:
                    # 'testPath' led to a non-existant node,
                    if infoHost:
                        log.err(
                            '%s does not exist in "%s:%s"\n'
                            % (shortPathToNode, infoHost, pathToInfo)
                            + "Aborting this build."
                        )
                    else:
                        log.err(
                            '%s does not exist in "%s"\n'
                            % (shortPathToNode, pathToInfo)
                            + "Aborting this build."
                        )
                    continue

                # else
                infoData = "\n".join(
                    [line.strip() for line in thisNode.text if len(line.strip()) > 0]
                )

                if len(infoData) == 0:
                    if infoHost:
                        log.err(
                            '%s exists in "%s:%s", but contains no info data.\n'
                            % (shortPathToNode, infoHost, pathToInfo)
                            + "Aborting this build."
                        )
                    else:
                        log.err(
                            '%s exists in "%s", but contains no info data.\n'
                            % (shortPathToNode, pathToInfo)
                            + "Aborting this build."
                        )
                    continue

                infoOptions = parser.parseLines(thisNode.text)
                log.stp('Parsed "%s"' % shortPathToNode)
                masterDict["testXmlNode"] = thisNode
            else:
                log.info('No "test.info" data provided')
                infoOptions = {}

            # determine 'buildDir', which will hold output from this individual build
            buildDir = os.path.normpath(testPath).replace(os.sep, "_")

            # determine absolute path to buildDir
            pathToBuildDir = os.path.join(pathToInvocationDir, buildDir)
            log.stp('Creating directory "%s"' % pathToBuildDir)

            if os.path.isdir(pathToBuildDir):
                # a directory of this name already exists[0]
                log.err(
                    'A directory called "%s" already exists.\n' % pathToBuildDir
                    + "Skipping this build."
                )
                continue

            # else
            try:
                os.mkdir(pathToBuildDir)
            except Exception as e:
                log.err("%s\n" % str(e) + "Skipping this build.")
                continue

            masterDict["pathToBuildDir"] = pathToBuildDir
            masterDict["buildDir"] = buildDir

            # enter override-options and "test.info" options into 'masterDict'
            # "test.info" options will temporarily overwrite override-options,
            # for the sake of constructing a good message to the user
            masterDict.update(overrideOptions)
            masterDict.update(infoOptions)

            # override-options take precedence over options with the same
            # name from the "test.info" file. Note in logfile if this happens.
            for key in overrideOptions:
                if key in masterDict and masterDict[key] != overrideOptions[key]:
                    log.note(
                        'key "%s" with value "%s" overridden by new value "%s"'
                        % (key, masterDict[key], overrideOptions[key])
                    )
                    masterDict[key] = overrideOptions[key]

            if testPath.count(os.sep) > 0:
                # Record the text of the original "test.info" data in the logfile
                log.info('****** "test.info" ******')
                log.info(infoData)
                log.info("*************************")

                # Record a copy of 'infoData' in this build's output directory.
                # The first line is an xml node-path to the original "test.info"
                # data, which FlashTestView uses to generate a link to the info-
                # file manager.
                msg = shortPathToNode + "\n\n"
                msg += infoData + "\n"
                open(os.path.join(pathToBuildDir, "test.info"), "w").write(msg)

            #####################
            ##  "errors" File  ##
            #####################

            # Creates a file "errors" which will record the number of
            # errors encountered for one build/series of runs as well
            # as the total number of parfiles intended to be run. The
            # file is a text file with 6 lines: one line each for the
            # number of errors in setup and compilation (will be 0 or
            # 1), one line each for the number of errors occurring in
            # runtime and in testing, a line for the total
            # number of parfiles that were to be tried in this build,
            # and finally a line for the number of errors in testing
            # that were "as before" (those are also included in the
            # number on the 4th line).
            startVals = "0\n0\n0\n0\n0\n0"
            open(os.path.join(pathToBuildDir, "errors"), "w").write(startVals)

            # The user can interfere again here if he/she wants, e.g.
            # to control what object is put in place for 'testObject's'
            # "setupper" and "compiler" members or to add an instruction
            # instruction to FlashTestView regarding what kind of a menu
            # item to generate on the "view builds" screen.
            testObject.entryPoint2()

            # Now that information from the "test.info" file has been
            # read in, add instances of strategy classes for setup,
            # compilation, execution, and testing stages.
            for keyword in ["setupper", "compiler"]:
                testObject.installComponent(keyword)

            #############
            ##  SETUP  ##
            #############
            log.info("** setup phase **")
            if testObject.setup() == False:
                __incErrors(0)
                continue

            ###################
            ##  COMPILATION  ##
            ###################
            log.info("** compilation phase **")
            if testObject.compile() == False:
                __incErrors(1)
                continue

            #################
            ##  TRANSFERS  ##
            #################
            transfers = masterDict.get("transfers", [])
            if len(transfers) > 0:
                # change space-delimited 'transfers' into a python list
                transfers = parser.stringToList(transfers)

            # make copies in output directory of any files in 'transfers'
            transferedFiles = []
            for path in transfers:
                # 'path' is assumed to be a relative path to the file
                # from the top-level directory of the FLASH source
                # Split this into path and basename
                path, file = os.path.split(path)
                log.info('Transfering file "%s"' % file)
                # HACK - this 'pathToFlash' stuff will be gone in next iteration
                pathToFlash = masterDict["pathToFlash"]
                if os.path.isfile(os.path.join(pathToFlash, path, file)):
                    copy(os.path.join(pathToFlash, path, file), pathToBuildDir)
                    transferedFiles.append(file)
                elif os.path.isdir(os.path.join(pathToFlash, path, file)):
                    os.symlink(
                        os.path.join(pathToFlash, path, file),
                        os.path.join(pathToBuildDir, file),
                    )
                    transferedFiles.append(file)
                else:
                    log.warn('File "%s" was not found in "%s"' % (file, path))

            # +-------------------+
            # |  EXTRA TRANSFERS  |
            # +-------------------+
            if masterDict.has_key("extraTransfers"):
                transfers = masterDict["extraTransfers"]
                if len(transfers) > 0:
                    # change space-delimited 'transfers' into a python list
                    transfers = parser.stringToList(transfers)

                    # make copies in output directory of any files in 'transfers' and append
                    # names to transferedFiles - if they exist.
                    for path in transfers:
                        # As above: 'path' is assumed to be a relative path to the file
                        # from the top-level directory of the FLASH source
                        # Split this into path and basename
                        path, file = os.path.split(path)
                        # HACK - this 'pathToFlash' stuff will be gone in next iteration
                        pathToFlash = masterDict["pathToFlash"]
                        if os.path.isfile(os.path.join(pathToFlash, path, file)):
                            #              log.info("os.path.join(pathToFlash=\"%s\",path=\"%s\",file=\"%s\") is \"%s\", pathToBuildDir is \"%s\".\n" %
                            #                       (pathToFlash, path, file, os.path.join(pathToFlash, path, file), pathToBuildDir))
                            copy(os.path.join(pathToFlash, path, file), pathToBuildDir)
                            transferedFiles.append(file)
                            log.info('Transferring extra file "%s"' % file)

            ##      log.info("transferedFiles is %s.\n" % transferedFiles)

            ################
            ##  PARFILES  ##
            ################
            parfiles = parser.stringToList(masterDict.get("parfiles", ""))
            masterDict["parfiles"] = parfiles

            if len(parfiles) > 0:
                # set number of parfiles to be run on fifth line of "errors" file
                errorsFile = os.path.join(pathToBuildDir, "errors")
                lines = open(errorsFile, "r").readlines()
                lines[4] = str(len(parfiles)) + "\n"
                open(errorsFile, "w").write("".join(lines))

            # DEV another entry point here for changing transfers and parfiles?

            #################
            ##  EXECUTION  ##
            #################
            if len(parfiles) > 0:
                log.info("** execution phase **")
            else:
                log.warn(
                    "Since parfiles is empty, there will be no execution or testing phases."
                )

            # inner loop over each parfile listed for a single executable
            for parfile in parfiles:

                if os.path.isabs(parfile):
                    if os.path.isfile(parfile):
                        parfileText = open(parfile).read()
                        parfile = os.path.basename(parfile)
                    else:
                        __incErrors(2)
                        log.err(
                            'Parfile "%s" not found\n' % parfile + "Skipping this run."
                        )
                        continue

                else:
                    __incErrors(2)
                    log.err(
                        'Parfiles specified in a "test.info" file such as "%s"\n'
                        % parfile
                        + "must be declared as absolute paths. Skipping this run."
                    )
                    continue

                # define the name of the directory that will hold all results from this run
                firstDot = parfile.find(".")
                if firstDot > 0:
                    runDir = parfile[:firstDot]  # truncate any dot and following chars
                    # (essentially to truncate the '.par')
                pathToRunDir = os.path.join(pathToBuildDir, runDir)

                if os.path.isdir(pathToRunDir):
                    # a directory of this name already exists in 'buildDir'
                    __incErrors(2)
                    log.err(
                        "There is already a directory called '%s' in %s\n"
                        % (runDir, pathToBuildDir)
                        + "Skipping this run."
                    )
                    continue
                # else all is well - The last level of nested directories will be
                # for the output of the run against this parfile
                os.mkdir(pathToRunDir)

                # set 'masterDict' layer to 2 and erase any keys previously set
                # at this layer, i.e. in an earlier iteration of the loop
                masterDict.setActiveLayer(2)

                masterDict["runDir"] = runDir
                masterDict["pathToRunDir"] = pathToRunDir

                # make links in 'runDir' to any transfers we had
                for transferedFile in transferedFiles:
                    try:
                        os.link(
                            os.path.join(pathToBuildDir, transferedFile),
                            os.path.join(pathToRunDir, transferedFile),
                        )
                    except OSError as e:
                        log.warn('Linking "%s" failed: %s' % (transferedFile, str(e)))

                # write the parfile into our new directory
                open(os.path.join(pathToRunDir, parfile), "w").write(parfileText)

                masterDict["pathToParfile"] = os.path.join(pathToRunDir, parfile)
                masterDict["parfile"] = parfile

                # create an "errors" file at the run level whose value will be
                # "00" (no errors), "10" (error during runtime) or "01" (error
                # during the testing phase)
                open(os.path.join(pathToRunDir, "errors"), "w").write("0\n0")

                # The user can interfere again here if he/she wants, e.g.
                # to control what object is put in place for 'testObject's'
                # "executer" and "tester" members
                testObject.entryPoint3()

                for keyword in ["executer", "tester"]:
                    testObject.installComponent(keyword)

                # run the executable
                if testObject.execute() == False:
                    __incErrors(2)
                    continue

                ############
                ##  TEST  ##
                ############
                log.info("** testing phase **")
                if testObject.test() == False:
                    __incErrors(3)

                # +-------------------------------------+
                # COPY COMPLETED TEST TO VIEW ARCHIVE  +
                # +-------------------------------------+

                # Check -t
    if "-vv" in flashTestOpts and "-t" not in flashTestOpts:
        # send to view archive

        if pathToViewArchive:
            log.stp("Sending copy of completed test output to view archive...")

            try:
                arch.sendToViewArchive(incremental=True)
            except Exception as e:
                log.err("%s\n" % e + "No copy of output sent to view archive.")
            else:
                log.stp("Incremental copy of output sent to view archive.")

    #############################
    ##  ARCHIVING AND CLEANUP  ##
    #############################

    log.stp("All tests completed.")
    # pathToInfo is real name, make this work remotely
    if masterInfoNode.dirty:
        log.stp("Info node has been modified, updating file.")
        try:

            def delta(old):
                remote = parseXmlString(old)
                remote.mergeDirtyFrom(masterInfoNode)
                return "".join(x + "\n" for x in remote.getXml())

            remoteFile.updatefile("%s:%s" % (infoHost, pathToInfo), delta)
        except Exception as e:
            log.err("Failed to update info node: " + str(e))

    if "-L" in flashTestOpts:
        if pathToLogArchive:
            log.stp("Sending log files to log archive...")
            try:
                arch.sendToLogArchive(False)
            except Exception as e:
                log.err("%s\n" % e + "No log files sent to log archive.")
            else:
                log.stp("Log files sent to log archive.")
        
    if "-t" in flashTestOpts:
        log.stp("No archiving done for test-run. FlashTest complete. End of Logfile.")
        os.remove(os.path.join(pathToInvocationDir, ".lock"))
    elif (not pathToMainArchive) and (not pathToViewArchive):
        log.stp("FlashTest complete. End of Logfile.")
        os.remove(os.path.join(pathToInvocationDir, ".lock"))
    else:
        log.stp(
            "Preparing data for archiving. Outcome of archiving attempt will be written to\n"
            + '"%s"\n' % os.path.join(pathToSiteDir, "archive.log")
            + "and will be incorporated into the regular logfile of the next invocation.\n"
            + "End of Logfile."
        )

        # instantiate archive logfile - this data will appear
        # in the next invocation's regular logfile
        archiveLog = Logfile(pathToSiteDir, "archive.log", "-v" in flashTestOpts)
        archiveLog.info("Archiving results for invocation %s:" % invocationDir, False)

        errorCreatingTarFile = False

        try:
            # file already removed
            os.remove(os.path.join(pathToInvocationDir, ".lock"))
        except OSError as e:
            if e.errno != os.errno.ENOENT:
                raise OSError(e)

        if pathToMainArchive:
            archiveLog.stp(
                'Creating local tarball "%s.tar.gz" for main archive...'
                % pathToInvocationDir
            )
            try:
                pathToTarFile = arch.makeTarFile()
            except Exception as e:
                errorCreatingTarFile = True
                archiveLog.err(
                    "Error creating tarball\n%s\n" % e
                    + "No files will be deleted from local copy."
                )
            else:
                try:
                    arch.sendTarFileToMainArchive(pathToTarFile)
                except Exception as e:
                    archiveLog.err(
                        "Unable to send tarball to main archive\n%s\n" % e
                        + 'Tarfile still exists at "%s"' % pathToTarFile
                    )
                else:
                    archiveLog.stp("Tarball sent to main archive.")
                    # remove local copy of tarball
                    os.remove(pathToTarFile)

                archiveLog.stp("Deleting specified files for slim copy...")
                # the files "deleted_files" that this method generates will obviously
                # only be present in the copy of the output sent to the view archive
                __deleteFiles(pathToInvocationDir)
                archiveLog.stp("Local files deleted.")

        if pathToViewArchive:
            if not pathToMainArchive or errorCreatingTarFile:
                archiveLog.stp("Sending fat copy of output to view archive...")
            else:
                archiveLog.stp("Sending slim copy of output to view archive...")

            try:
                open(os.path.join(pathToInvocationDir, ".lock"), "w").write("")
                arch.sendToViewArchive("-vv" in flashTestOpts)
            except Exception as e:
                archiveLog.err("%s\n" % e + "No copy of output sent to view archive.")
            else:
                if not pathToMainArchive or errorCreatingTarFile:
                    archiveLog.stp("Fat copy of output sent to view archive.")
                else:
                    archiveLog.stp("Slim copy of output sent to view archive.")

            os.remove(os.path.join(pathToInvocationDir, ".lock"))

        archiveLog.stp("FlashTest complete.")

    pidFile.unregisterPid(pathToPidFile)


def usage():
    print('Usage for "flashTest.py":')
    print("  ./flashTest.py [general opts] [test-path#1] [override opts] \\")
    print("                                [test-path#2] [override opts] \\")
    print("                                [test-path#n] [override opts]")
    print("or:")
    print("  ./flashTest.py [general opts] -f [path/to/job/file] \\")
    print("                                   [test-path#1][override opts]")
    print("")
    print("General options to FlashTest are:")
    print('-c <file>: use <file> as "config" instead of default')
    print('-i <file>: use <file> as source of "test.info" data')
    print('-f <file>: read <file> (a "job-file") for list of test-paths')
    print("-o <dir> : direct output to <dir>")
    print("-s <name>: <name> is the name of this site (host)")
    print('-e <file>: <file> is used as the "exeScript" instead of the default')
    print("-t       : test run (no archiving)")
    print("-u       : update FLASH before run")
    print("-v       : verbose output (print logfile as it is written)")
    print(
        "-vv      : send results to view archive incrementally (only if -t is not used)"
    )
    print("-L       : send log files to log archive")
    print("-z <dir> : FLASH source rooted at <dir>")
    print(
        "--force  : Run test even if other instance(s) of flashTest currently running"
    )
    print("")
    print("Each option must be entered with its own preceding dash. For example")
    print("")
    print("  $ ./flashTest.py -t -v -f path/to/job/file")
    print("")
    print("will work, but")
    print("")
    print("  $ ./flashTest.py -tvf path/to/job/file")
    print("")
    print("will not.")
    print("")
    print("-i, -o, and -z options will override 'pathToInfo', 'pathToOutdir', and")
    print("'pathToFlash' in the \"config\" file.")
    print("")
    print('Variables that are set in the "config" file or in a "test.info" file')
    print("may be overridden on the command line, or in a job-file, by so-called")
    print('"override options" e.g.:')
    print("")
    print("  $ ./flashTest.py Comparison/Sod key1=val1 key2=val2")
    print("")
    print("If an option passed in this way, for example, 'key1', requires no value")
    print("the syntax looks like:")
    print("")
    print("  $ ./flashTest.py Comparison/Sod key1= key2=val2")
    print("")
    print("Different entryPoint, setupper, compiler, executer, and tester components")
    print('can be assigned to different test-species in the "config" file. See the')
    print('notes in "config" for more details or see the FlashTest User\'s Guide.')


if __name__ == "__main__":
    # add "lib" to our sys.path. It will be in the
    # same directory where the executable is found
    sys.path.insert(
        0, os.path.join(os.path.abspath(os.path.dirname(sys.argv[0])), "lib")
    )
    import flashTestParser as parser
    from archive import Archive
    from getProcessResults import getProcessResults
    from layeredDict import LayeredDict
    from logfile import Logfile
    from testObject import TestObject
    from xmlNode import parseXml, parseXmlString
    import pidFile
    import remoteFile

    locale.setlocale(locale.LC_CTYPE, '')

    if len(sys.argv) == 1 or sys.argv[1] == "-h":
        usage()
    else:
        main()




#!/usr/bin/env python3
import sys, re, os, shutil

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

# regular expressions for dirs we won't descend into
# to search for logfiles
avoidPatterns = ["\.svn", "2005-09-*", "2005-10-*", "2005-11-*"]

def matchesAvoidPattern(item):
  for avoidPattern in avoidPatterns:
    if re.match(avoidPattern, item):
      return True
  return False

def main():

  scriptDir = os.path.abspath(os.path.dirname(sys.argv[0]))
  user = "tester"
  experiment = "test-suite runs"
  template = ("user: %s\n" +
              "experiment: %s\n" +
              "grouptag: %s\n")

  results = []

  def __createDotUploadFiles(dir):
    cwd = os.getcwd()
    os.chdir(dir)
    items = os.listdir(".")
    for item in items:
      if item.endswith(".log"):
        # see if this logfile is from a successful run
        if os.path.isfile("errors"):
          errors = open("errors","r").read()
          if "1" in errors:
            continue
          # else all "0"s in "errors" means this was a successful run
          runName   = os.getcwd()
          buildName = os.path.dirname(runName)
          dateName  = os.path.dirname(buildName) # we're going to ignore this one
          siteName  = os.path.dirname(dateName)

          runName   = os.path.basename(runName)
          buildName = os.path.basename(buildName)
          dateName  = os.path.basename(dateName)
          siteName  = os.path.basename(siteName)
          groupTag  = siteName + "_" + buildName + "_" + runName
          try:
            outdir = os.path.join("/home/tester/flashTest/outputcopy",siteName,dateName,buildName,runName)
            if not os.path.isdir(outdir):
              os.makedirs(outdir)
              shutil.copy(item, outdir)
              open(os.path.join(outdir,".upload"),"w").write(template % (user, experiment, groupTag))
              results.append("wrote grouptag:\n\t%s\nin dir:\n\t%s" % (groupTag, outdir))
          except:
            results.append("error writing grouptag:\n\t%s\nin dir:\n\t%s" % (groupTag, outdir))
        else:
          # "errors" file is suspiciously not here...
          continue
      elif os.path.isdir(item) and not matchesAvoidPattern(item):
        __createDotUploadFiles(item)
    os.chdir(cwd)

  if sys.argv[1] == "-d":
    if len(sys.argv) != 3:
      usage()
    else:
      flashTestDir = sys.argv[2]
      cwd = os.getcwd()
      pathToFlashTestDir = os.path.join(cwd, flashTestDir)
      if not os.path.isdir(pathToFlashTestDir):
        print("error: no such directory: %s" % pathToFlashTestDir)
        sys.exit(1)
      # else we know flashTestDir is really a directory
  else:
    flashTestDir = scriptDir
  
  __createDotUploadFiles(flashTestDir)
  for result in results:
    print(result)
  
def usage():
  print("Do this later")

if __name__ == '__main__':
  if len(sys.argv) == 1 or sys.argv[1] == "-h":
    usage()
  else:
    main()



"""Python API for FlashXTest"""

import os
from .. import lib


def init(**apiDict):
    """
    Initialize test configuration

    Arguments
    ---------
    apiDict : Dictionary to populate Config file
    """
    # Cache the value to current directory and set it as
    # testDir in apiDict
    apiDict["testDir"] = os.getcwd()

    # Set Config file
    __setConfig(apiDict)

    # Set exeScript
    __setExeScript(apiDict)


def __setExeScript(apiDict):
    """
    Arguments:
    ---------
    apiDict: Dictionary to populate Config file
    """
    apiDict["pathToExeScript"] = apiDict["testDir"] + "/exeScript"
    lib.config.setExe(apiDict)


def __setConfig(apiDict):
    """
    Arguments:
    ---------
    apiDict: Dictionary to populate Config file
    """
    # Cache the value of user Config file and store it as
    # pathToConfig in apiDict
    apiDict["pathToConfig"] = apiDict["testDir"] + "/config"

    # Check if pathToConfig already exists and
    # skip the setup process
    if os.path.exists(apiDict["pathToConfig"]):
        print(
            lib.colors.WARNING
            + "[FlashXTest] Skipping initialization: Config file already exists!"
        )

    # Setup configuration if pathToConfig does not exist
    else:
        lib.config.setConfig(apiDict)



import re, os

class EntryPointTemplate:
  def __init__(self, testObject):
    self.owner      = testObject
    self.masterDict = testObject.masterDict

  def entryPoint1(self):
    return True

  def entryPoint2(self):
    return True

  def entryPoint3(self):
    return True


class SetupperTemplate:
  def __init__(self, testObject):
    self.owner = testObject
    self.masterDict = testObject.masterDict

  def setup(self):
    log = self.masterDict["log"]  # guaranteed to exist by flashTest.py
    #log.err("No implementation provided for method \"setup\".")
    #return False
    log.stp("Nothing to setup.")
    return True


class CompilerTemplate:
  def __init__(self, testObject):
    self.owner = testObject
    self.masterDict = testObject.masterDict

  def compile(self):
    log = self.masterDict["log"]  # guaranteed to exist by flashTest.py
    #log.err("No implementation provided for method \"compile\".")
    #return False
    log.stp("Nothing to compile.")
    return True

  def getDeletePatterns(self):
    return ["delete_me-"]

  def adjustFilesToDelete(self, filesToDelete):
    pass

  def writeFilesToDelete(self):
    """
    Gets a list of regular expressions from "getDeletePatterns()" and tries
    to match them against the files found in 'pathToBuildDir'. If any matches
    are found, the list is submitted to 'adjustFilesToDelete()' for possible
    modification. Afterwards, any elements remaining in 'filesToDelete' will
    be written into the text file "files_to_delete". The files indicated in
    this file will be deleted after the invocation is complete.

    The user may use 'adjustFilesToDelete' to remove from the list any files
    that would normally be deleted.
    """
    pathToBuildDir = self.masterDict["pathToBuildDir"]  # guaranteed to exist by flashTest.py

    deletePatterns = self.getDeletePatterns()

    filesToDelete = []
    files = [item for item in os.listdir(pathToBuildDir) if os.path.isfile(os.path.join(pathToBuildDir, item))]
    for file in files:
      for deletePattern in deletePatterns:
        if re.match(deletePattern, file):
          filesToDelete.append(file)

    if filesToDelete:
      filesToDelete.sort()
      self.adjustFilesToDelete(filesToDelete)

    if filesToDelete:
      open(os.path.join(pathToBuildDir, "files_to_delete"),"w").write("\n".join(filesToDelete))


class ExecuterTemplate:
  def __init__(self, testObject):
    self.owner = testObject
    self.masterDict = testObject.masterDict

  def execute(self):
    log = self.masterDict["log"]  # guaranteed to exist by flashTest.py
    #log.err("No implementation provided for method \"execute\".")
    #return False
    log.stp("Nothing to execute.")
    return True

  def getDeletePatterns(self):
    return ["delete_me-"]

  def adjustFilesToDelete(self, filesToDelete):
    pass

  def writeFilesToDelete(self):
    """
    Gets a list of regular expressions from "getDeletePatterns()" and tries
    to match them against the files found in 'pathToRunDir'. If any matches
    are found, the list is submitted to 'adjustFilesToDelete()' for possible
    modification. Afterwards, any elements remaining in 'filesToDelete' will
    be written into the text file "files_to_delete". The files indicated in
    this file will be deleted after the invocation is complete.

    The user may use 'adjustFilesToDelete' to remove from the list any files
    that would normally be deleted.
    """
    pathToRunDir = self.masterDict["pathToRunDir"]  # guaranteed to exist by flashTest.py

    deletePatterns = self.getDeletePatterns()

    filesToDelete = []
    files = [item for item in os.listdir(pathToRunDir) if os.path.isfile(os.path.join(pathToRunDir, item))]
    for file in files:
      for deletePattern in deletePatterns:
        if re.match(deletePattern, file):
          filesToDelete.append(file)

    if filesToDelete:
      filesToDelete.sort()
      self.adjustFilesToDelete(filesToDelete)

    if filesToDelete:
      open(os.path.join(pathToRunDir, "files_to_delete"),"w").write("\n".join(filesToDelete))


class TesterTemplate:
  def __init__(self, testObject):
    self.owner      = testObject
    self.masterDict = testObject.masterDict

  def openOutfile(self):
    pathToRunDir    = self.masterDict["pathToRunDir"]     # guaranteed to exist by flashTest.py
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py
    
    os.chdir(pathToRunDir)
    outfile = open("test_output","w")
    self.masterDict["outfile"] = outfile

    os.chdir(pathToRunDir)

  def test(self):
    log = self.masterDict["log"]  # guaranteed to exist by flashTest.py
    #log.err("No implementation provided for method \"test\".")
    #return False
    log.stp("Nothing to test.")
    return True

  def closeOutfile(self):
    pathToFlashTest = self.masterDict["pathToFlashTest"]  # guaranteed to exist by flashTest.py
    outfile = self.masterDict["outfile"]                  # guaranteed to exist by openOutfile() (above)

    self.masterDict["outfile"].close()

    os.chdir(pathToFlashTest)



import os, re, sys

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

class ParseError(Exception):
  pass

def __cleanLines(lines):

  l = 0
  while l < len(lines):
    c = 0
    while c < len(lines[l]):
      if lines[l][c] == '"':
        # jump forward to matching double-quote
        c = lines[l].find('"',c+1)
      elif lines[l][c] == "'":
        # jump forward to matching single-quote
        c = lines[l].find("'",c+1)
      elif lines[l][c] == "#":
        # trim off any comments (not btwn quotes)
        lines[l] = lines[l][:c]

      if c < 0:
        # one of the above finds came up empty
        # because of an unmatched quote mark
        break
      else:
        c += 1
    l += 1

  # remove blank lines
  lines = [line.strip() for line in lines if len(line.strip()) > 0]

  return lines


def fileToList(pathToFile):
  """
  Open 'filename' and parse its text into a list. The elements of
  the list will be the lines of the text as delimited by newlines.
  Comments and blank lines are elimnated in all cases.

  This function is called by "flashTest.py" to get lists of lines
  from "errors" and "files_to_delete" files, and also to turn user
  test-path files (such as follow the "-f" option to flashTest.py)
  into lists of arguments.
  """
  lines = open(pathToFile).read().split("\n")
  return __cleanLines(lines)


def getPathsAndOpts(args):

  def __linkEqualsSigns(args):
    """
    Join up any dangling equals signs, e.g.:

      test/path bar =baz   -->   test/path bar=baz

    such that 'bar=baz' becomes an option to test-path.
    Note, however, that the dangling equals in:

      foo bar= baz
 
    is not joined to 'baz', as this syntax indicates that
    'bar' is an option to test-path that takes no argument
    """
    args2 = []
    lastArg = None
    for arg in args:
      if lastArg and arg.startswith("="):
        if (lastArg + arg).count("=") > 1:
          raise ParseError("Options to tests must be of form 'key=value', not '%s %s'" % (lastArg, arg))
        else:
          args2[-1] = args2[-1] + arg
      else:
        args2.append(arg)
      lastArg = arg
    return args2

  pathsAndOpts = []

  if args:
    args = __linkEqualsSigns(args)
    if args[0].startswith("="):
      raise ParseError("test-option value \"%s\" has no corresponding key." % args[0])
    # else
    if args[0].count("=") > 0:
      raise ParseError("option \"%s\" must follow a test-path" % args[0])
    # else

    i = 0
    while i < len(args):
      if args[i].startswith("-"):
        msg = ("The option \"%s\" is not allowed in this position.\n" % args[i] +
               "Options to flashTest.py must precede all test-paths.")
        raise ParseError(msg)
      # else
      if args[i].count("=") >= 1:
        key, val = args[i].split("=",1)
        pathsAndOpts[-1][1][key] = val.strip("'\"")
      else:
        pathsAndOpts.append((args[i], {}))
      i+=1

  return pathsAndOpts


def parseCommandLine(args, standAloneOpts=None):
  flashTestOpts = {}

  # Options to flashTest.py must come immediately after the name of the
  # executable itself and before any paths or options to specific tests.
  i = 0
  while i < len(args):
    if not args[i].startswith("-"):
      break
    # else, this is an option
    if args[i] in standAloneOpts:
      # this option takes no argument
      flashTestOpts[args[i]] = ""
    else:
      # the option's argument is the next element in the list
      if i+1 == len(args):
        # but there are no more elements in the list!
        raise ParseError("Option %s requires an argument" % args[i])
      else:
        flashTestOpts[args[i]] = args[i+1]
        i += 1
    i += 1

  # Now, everything left in 'args' should be paths to tests and
  # command-line arguments to those tests.
  pathsAndOpts = getPathsAndOpts(args[i:])

  return (flashTestOpts, pathsAndOpts)


def parseFile(pathToFile):
  """
  Take a path to a textfile of newline-delimited key/value pairs and
  return a corresponding dictionary.

  key/value pairs are denoted by:

    key: value

  section headers are given as:

    [header]

  The header will be prepended to any keys found below it so:

    [header]
    foo: fooval

  will yield {"header.foo": "fooval"} in 'returnDict'
  """
  lines = open(pathToFile).read().split("\n")
  return parseLines(lines)


def parseLines(lines):
  """
  Take some text of newline-delimited key/value pairs
  and return a corresponding dictionary
  """
  returnDict = {}
  header = ""

  lines = __cleanLines(lines)

  for line in lines:
    if re.search("^\S+?:", line):
      key = line.split(":",1)[0].strip()
      val = line.split(":",1)[1].strip()
      if header:
        key = "%s.%s" % (header, key)
      returnDict[key] = val
    elif re.search("^\[.*\]$", line):
      header = line.strip("[]").strip()

  return returnDict


def stringToList(s):
  """
  convenience method for converting 'parfiles' and 'transfers'
  strings into lists for use by flashTest.py
  """
  if len(s.strip()) > 0:
    return re.split("\s+",s.strip())
  else:
    return []



import sys,os

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

import strategyTemplates

class TestObject:
  """
  Represents a single object whose public methods define behavior for all
  phases of a FlashTest run. TestObject is never sub-classed. Instead, each
  instance is assembled out of components which are themselves instances of
  user-defined strategy classes specified by the user in the "config" file
  and in "test.info" files.

  For example, the user may specify in the "config" file that all calls to
  "setup()" should be calls to the setup method of the user-defined class
  "MySetupper", that calls to "test()" for problems of type "Comparison"
  should be calls to the test method of the user-defined class "MyTester",
  and that calls to "test()" for problems of type "UnitTestUG" should be
  calls to the test method of the user-defined class "MyTester2".

  Here is the relevant snippet of our example "config" file:

    setupper: MySetupper

    [Comparison]
    tester: MyTester

    [UnitTestUG]
    tester: MyTester2

  However, the user may override the values in "config" with values from a
  "test.info" file. e.g., if a subset of tests of type "UnitTestUG" require
  the test method of the user-defined class MyTester3, the file:

  flashTest/infoFiles/<mySite>/UnitTestUG/<someSubset>/test.info

  can define:

    tester: MyTester3


  Components not specified by the user in "config" or "test.info" files are
  instantiated from templates defined in "strategyTemplates.py".

  Components which are specified by the user should subclass these same
  templates.

  TestObject instances define a component "entryPoint" on initialization. If
  the user wishes to specify a custom entry-point class, he or she must do
  so in the "config" file. This single component can not be overridden in a
  "test.info" file.

  After initialization, flashTest.py will call the method installComponent()
  for the setupper, compiler, executer, and tester components.
  """
  def __init__(self, masterDict):
    self.entryPoint = None
    self.setupper   = None
    self.compiler   = None
    self.executer   = None
    self.tester     = None
    self.masterDict = masterDict
    # give this object an entryPoint object
    self.installComponent("entryPoint")


  def installComponent(self, component):
    """
    called first by TestObject's __init__() method and later by flashTest.py to
    instantiate a user-defined strategy class or, if this is not specified, of
    a default strategy template.
    """
    if self.masterDict.has_key(component):
      className = self.masterDict[component]
      # 'className' is the name of the strategy class that the user
      # wants to instantiate as the "setupper", "compiler", etc, but
      # we still need to know what module it's in...

      # Does 'className' specify its module by the form "myModule.MySetupper"?
      if className.find(".") > 0:
        modName, className = className.split(".",1)
      # No? Then has the user specified a "useModule" key?
      elif self.masterDict.has_key("useModule"):
        modName = self.masterDict["useModule"]
      # No? Then we don't know what module to look in.
      else:
        errMsg = ("You must indicate a module where class \"%s\" can be found.\n" % className +
                  "Add a \"useModule\" key to \"config\" or indicate the module\n" +
                  "with \"modName.%s\"\n" % className +
                  "Modules should be located in the FlashTest \"lib\" directory.")
        raise Exception(errMsg)
      # At this point, 'modName' has been defined, so we know where to look for 'className'
      if modName not in sys.modules:
        __import__(modName)

    else:
      modName = "strategyTemplates"
      # Name of default setupper class is "SetupperTemplate",
      # default compiler is "CompilerTemplate", etc.
      className = "%sTemplate" % (component[0].upper() + component[1:])


    # Check to see if the requested component from
    # the requested module is already in place
    if (not hasattr(self, component) or
        getattr(self, component).__class__.__name__ != className or
        getattr(self, component).__module__ != modName):

      # intantiate 'className' and make 'instance' an attribute of 'self'
      if hasattr(sys.modules[modName], className):
        instance = getattr(sys.modules[modName], className)(self)
        setattr(self, component, instance)
      else:
        errMsg = ("Module \"%s\" defines no class called \"%s\"\n" % (modName, className))
        raise Exception(errMsg)


  # These methods are called by flashTest.py and invoke a method
  # or methods on the instance of the strategy class specified in
  # "config" and instantiated in TestObject.__init__ (above)
  def entryPoint1(self):
    return self.entryPoint.entryPoint1()

  def entryPoint2(self):
    return self.entryPoint.entryPoint2()

  def entryPoint3(self):
    return self.entryPoint.entryPoint3()

  def setup(self):
    return self.setupper.setup()

  def compile(self):
    result = self.compiler.compile()
    self.compiler.writeFilesToDelete()
    return result

  def execute(self):
    result = self.executer.execute()
    self.executer.writeFilesToDelete()
    return result

  def test(self):
    self.tester.openOutfile()
    result = self.tester.test()
    self.tester.closeOutfile()
    return result



"""FlashXTest library to interface with backend.FlashTest"""

import os, sys, subprocess
import warnings
import yaml

from .. import backend
from .. import lib


class __YamlLoader(yaml.SafeLoader):
    """
    Class YamlLoader for YAML
    """

    def __init__(self, stream):
        """
        Constructor
        """
        super().__init__(stream)
        self._stream = stream

    def construct_mapping(self, node, deep=False):
        """
        Mapping function
        """
        mapping = set()
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise ValueError(
                    lib.colors.FAIL
                    + f"[FlashXTest] Duplicate {key!r} key found in {self._stream.name!r}."
                )
            mapping.add(key)
        return super().construct_mapping(node, deep)


def parseYaml(mainDict, setupName):
    """
    Arguments:
    ----------
    mainDict  : Main dictionary
    setupName : Setup name
    """
    # Get path to simulation directory
    yamlFile = (
        mainDict["pathToFlash"]
        + "/source/Simulation/SimulationMain/"
        + setupName
        + "/tests/tests.yaml"
    )

    with open(yamlFile, "r") as stream:
        try:
            yamlDict = yaml.load(stream, Loader=__YamlLoader)
        except yaml.YAMLError as exc:
            print(exc)

    for nodeName in yamlDict.keys():
        for key in yamlDict[nodeName].keys():
            if key not in [
                "setupOptions",
                "parfiles",
                "restartParfiles",
                "transfers",
            ]:
                raise ValueError(
                    lib.colors.FAIL
                    + f'[FlashXTest] unrecognized key "{key}" for "{nodeName}" '
                    + f"in {yamlFile}"
                )

    return yamlDict



"""Build and installation script for flashxtest."""

# standard libraries
import re
from setuptools import setup, find_packages

# get long description from README
with open("README.rst", mode="r") as readme:
    long_description = readme.read()

with open("FlashXTest/__meta__.py", mode="r") as source:
    content = source.read().strip()
    metadata = {
        key: re.search(key + r'\s*=\s*[\'"]([^\'"]*)[\'"]', content).group(1)
        for key in ["__pkgname__", "__version__", "__authors__", "__license__"]
    }

# core dependencies - click, docker, singularity
DEPENDENCIES = ["click", "six", "argparse", "pyyaml"]

setup(
    name=metadata["__pkgname__"],
    version=metadata["__version__"],
    author=metadata["__authors__"],
    license=metadata["__license__"],
    packages=find_packages(where="./"),
    package_dir={"": "./"},
    package_data={
        "": [
            "backend/FlashTest/ERROR",
            "backend/FlashTest/error.sh",
            "backend/FlashTest/exeScript",
            "backend/FlashTest/configBase",
            "backend/FlashTest/configTemplate",
            "backend/FlashTest/scripts/changeInfoFiles",
            "backend/FlashTest/scripts/restarttest/restarttest.sh",
        ]
    },
    scripts=["FlashXTest/flashxtest"],
    include_package_data=True,
    long_description=long_description,
    classifiers=["Programming Language :: Python :: 3.8"],
    install_requires=DEPENDENCIES,
)



#!/usr/bin/env python3

import os, errno, os.path
from fcntl import LOCK_UN, LOCK_EX,LOCK_NB, flock
import six

import struct
#MAXPIDBYTE=struct.calcsize("P")
MAXPIDBYTE=8
MAXPID=65536


def lockFile(fd,nonBlock=True):
        try:
                if nonBlock:
                        flock(fd,LOCK_EX|LOCK_NB)
                else:
                        flock(fd,LOCK_EX)
        except Exception as e:
                return False
        return True

def unlockFile(fd):
        flock(fd, LOCK_UN)


def unregisterPid(pidFile):
                print("trying to delete:", pidFile)
                pid=os.getpid()
                try:
                        #Open file, obtain lock and remove if this flashTest owns it
                        #This will block. Not sure O_EXCL is not ignored
                        fd= os.open(pidFile, os.O_RDONLY|os.O_EXCL)
                        print(fd)
                        if lockFile(fd, False):
                                apid = os.read(fd,MAXPIDBYTE)
                                print(apid.strip())
                                apid=struct.unpack('>HHHH',apid)[0]
                                #print(type(pid), "new ", type(apid.strip()))
                                if str(pid) != apid:
                                        unlockFile(fd)
                                        os.close(fd)
                                        print("deleting finally")
                                        os.remove(pidFile)
                                else:
                                        print("else block")
                                        unlockFile(fd)
                                        os.close(fd)
                except Exception as e:
                                print("Unable to delete my own /tmp/flashTest.pid")
                                pass
def main():
    pidFile = "/tmp/flashTest.pid"
    unregisterPid(pidFile)
main()



"""FlashXTest library to interface with backend.FlashTest"""

import os, subprocess

from .. import backend
from .. import lib


def getMainDict(apiDict):
    """
    Arguments
    --------
    apiDict  : Dictionary to override values from Config file

    Returns
    -------
    mainDict: Dictionary for keys in the config file
    """
    # Build Config file for mainDict.
    # Read the user Config file (configApi), append it to Base Config from backend (configBase),
    # and create a new Config (configMain) in 'testDir/.fxt' folder
    configApi = apiDict["pathToConfig"]
    configMain = configApi

    # Parse the configMain file
    mainDict = backend.flashTestParser.parseFile(configMain)

    # Update mainDict with values from apiDict
    mainDict.update(apiDict)

    return mainDict


def setExe(apiDict):
    """
    Arguments
    ---------
    apiDict : API dictionary
    """
    pass


def setConfig(apiDict):
    """
    Setup configuration

    Arguments
    ---------
    apiDict    : API dictionary
    """
    # Get path to configuration template from FlashTest backend
    configTemplate = os.path.dirname(backend.__file__) + "/FlashTest/configTemplate"

    # Get path to configuration base from FlashTest backend
    configBase = os.path.dirname(backend.__file__) + "/FlashTest/configBase"

    # Get path to user configuration file from apiDict
    configFile = apiDict["pathToConfig"]

    # Start building configFile from configTemplate
    #
    # configTemplate in read mode as ctemplate
    # configFile in write mode as cfile
    #
    with open(configTemplate, "r") as ctemplate, open(configFile, "w") as cfile:

        # Read lines from ctemplate
        lines = ctemplate.readlines()

        # Iterate over lines and set values defined in apiDict
        for line in lines:

            # Set path to Archive
            line = line.replace(
                "pathToLocalArchive:",
                str("pathToLocalArchive: " + apiDict["testDir"] + "/TestArchive"),
            )

            # Set default baseLineDir
            line = line.replace(
                "baselineDir:",
                str("baselineDir:        " + apiDict["testDir"] + "/TestArchive"),
            )

            # Set default pathToOutdir
            line = line.replace(
                "pathToOutdir:",
                str("pathToOutdir:       " + apiDict["testDir"] + "/TestResults"),
            )

            # Set 'pathToFlash' if defined in apiDict
            if "pathToFlash" in apiDict:
                line = line.replace(
                    "pathToFlash:",
                    str("pathToFlash:        " + str(apiDict["pathToFlash"])),
                )

            # Set 'flashSite' if define in apiDict
            if "flashSite" in apiDict:
                line = line.replace(
                    "flashSite:",
                    str("flashSite:          " + str(apiDict["flashSite"])),
                )

            cfile.write(line)

    # Append additional options from configBase
    #
    with open(configBase, "r") as cbase, open(configFile, "a") as cfile:
        cfile.write("\n")
        cfile.write("# Following options are default values that should\n")
        cfile.write("# not be changed for most cases \n")

        lines = cbase.readlines()

        for line in lines:
            cfile.write(line)

    print(lib.colors.OKGREEN + "[FlashXTest] Initialized configuration")



"""Python CLI for flashxtest"""

import os
import subprocess
import click
import pkg_resources

from .. import api


@click.group(name="flashxtest", invoke_without_command=True)
@click.pass_context
@click.option("--version", "-v", is_flag=True)
def flashxtest(ctx, version):
    """
    \b
    Command line interface for managing
    Flash-X testing framework. Type --help
    for individual commands to learn more.
    """
    if ctx.invoked_subcommand is None and not version:
        subprocess.run(
            f"export PATH=~/.local/bin:/usr/local/bin:$PATH && flashxtest --help",
            shell=True,
            check=True,
        )

    if version:
        click.echo(pkg_resources.require("FlashXTest")[0].version)


@flashxtest.command(name="init")
@click.option("--source", "-z", default=None, help="Flash-X source directory")
@click.option("--site", "-s", default=None, help="Flash-X site name")
def init(source, site):
    """
    \b
    Initialize site specific configuration.

    \b
    This command is used to setup site specific
    configuration for your testing environment
    using "config" and "execScript" files. At
    present only "config" file is created since
    this feature is under development.

    \b
    This command will not work if "config" file
    is present in the working directory. To edit
    existing site specific configuration, "config"
    should be edited directly.
    """
    # Arguments
    # ---------
    # source: Flash-X source directory
    # site: Flash-X site name
    api.init(flashSite=site, pathToFlash=source)


@flashxtest.command(name="setup-suite")
@click.argument("suitelist", type=str, nargs=-1)
def setup_suite(suitelist):
    """
    \b
    Create a "test.info" from a list of suite files.

    \b
    This command accepts multiple files with suffix,
    ".suite" to build a "test.info". If no arguments are
    supplied, all "*.suite" files are used from the working
    directory.

    \b
    The ".suite" files represent a collection of mutually
    exclusive test specifications associated with a "config" file.
    Each line in a file represent a unique test specification
    defined as,

    \b
    incompFlow/LidDrivenCavity -t "UnitTest/LidDrivenCavity/AMReX/2d" -np 4 --debug

    \b
    The first value represents a Flash-X setup defined in
    source/Simulation/SimulationMain directory with following
    options,

    \b
    -t, --test TEXT	Defined in */tests/tests.yaml)
    -np, --nprocs TEXT	Number of processors
    --debug BOOLEAN	Debug test

    """
    api.setup_suite(pathToSuites=suitelist)


@flashxtest.command(name="run")
def run():
    """
    \b
    Run the test suite using "test.info".

    \b
    This command runs all the tests defined in
    "test.info", and conveys errors
    """
    # Arguments
    # ---------
    # testsuite : string for the test suite file
    api.run()



"""FlashXTest library to interface with backend.FlashTest"""

import os, subprocess

from .. import backend
from .. import lib


def specListFromNode(infoNode, specList):
    """
    Create a list of node paths by recursively searching
    till the end of the tree

    Arguments
    ---------
    infoNode  : FlashTest node object
    specList  : Empty list for test specifications
    """
    if infoNode.subNodes:
        for subNode in infoNode.subNodes:
            specListFromNode(subNode, specList)
    else:
        xmlDict = {}
        for xmlEntry in infoNode.text:
            key, value = xmlEntry.split(":")
            xmlDict.update({key: value.strip()})

        xmlDict["nodeName"] = infoNode.getPathBelowRoot()

        specList.append(xmlDict)


def jobListFromNode(infoNode, jobList):
    """
    Create a list of node paths by recursively searching
    till the end of the tree

    Arguments
    ---------
    infoNode : FlashTest node object
    jobList  : Empty jobList
    """
    if infoNode.subNodes:
        for subNode in infoNode.subNodes:
            jobListFromNode(subNode, jobList)
    else:
        jobList.append(infoNode.getPathBelowRoot())


def addNodeFromPath(infoNode, nodePath):
    """
    infoNode : node object
    nodePath : node path
    """
    nodeList = nodePath.split("/")

    for node in nodeList:
        if not infoNode.findChild(node):
            infoNode.add(node)

        infoNode = infoNode.findChild(node)


def createInfo(mainDict, specList):
    """
    Get test info site

    Arguments:
    -----------
    mainDict : Main dictionary
    specList : List of test specifications
    """
    # Set variables for site Info
    pathToInfo = str(mainDict["testDir"]) + "/test.info"

    if os.path.exists(pathToInfo):
        overwrite = input(
            lib.colors.WARNING
            + f"[FlashXTest] {pathToInfo!r} already exits. Replace? (Y/n) "
        )

        if overwrite == "y" or overwrite == "Y":
            print(lib.colors.OKGREEN + "OVERWRITING")
        else:
            print(lib.colors.OKGREEN + "SKIPPING")
            return

    # Get uniquie setup names
    setupList = []
    for testSpec in specList:
        setupList.append(testSpec.setupName)
    setupList = [*set(setupList)]

    # Get yaml dictionary
    setupYaml = {}
    for setupName in setupList:
        setupYaml[setupName] = lib.yml.parseYaml(mainDict, setupName)

    # Build test.info file from the test suite
    with open(pathToInfo, "w") as testInfoFile:

        # Create xml node to store info
        infoNode = backend.FlashTest.lib.xmlNode.XmlNode("infoNode")

        # Add the site node
        infoNode.add(f'{mainDict["flashSite"]}')

        # create test node from suiteDict
        for testSpec in specList:

            addNodeFromPath(
                infoNode.findChild(f'{mainDict["flashSite"]}'), testSpec.nodeName
            )

            setupInfo = setupYaml[testSpec.setupName][testSpec.nodeName]

            for key in setupInfo.keys():
                if hasattr(testSpec, key):
                    setattr(testSpec, key, setupInfo[key])
                else:
                    raise ValueError(
                        f"{key!r} defined for test {testSpec.nodeName!r}"
                        + f" in {testSpec.setupName!r} does exist in TestSpec"
                    )

            infoNode.findChildrenWithPath(testSpec.nodeName)[
                0
            ].text = testSpec.getXmlText()

        # Write xml to file
        for line in infoNode.getXml():
            testInfoFile.write(f"{line}\n")

    mainDict["pathToInfo"] = pathToInfo

    print(lib.colors.OKGREEN + "[FlashXText] test.info is setup")



#!/usr/bin/env python3

import os, errno, os.path, sys
from fcntl import LOCK_UN, LOCK_EX,LOCK_NB, flock 

sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))

MAXPIDBYTE=8
MAXPID=65536


def lockFile(fd,nonBlock=True):
        try:
                if nonBlock:
                        flock(fd,LOCK_EX|LOCK_NB)
                else:
                        flock(fd,LOCK_EX)
        except Exception as e:
                return False
        return True

def unlockFile(fd):
        flock(fd, LOCK_UN)

def registerPid(pidFile, forceRun=False):
        pid = os.getpid()
        try:
                fd=os.open(pidFile, os.O_CREAT|os.O_WRONLY|os.O_EXCL)
                if lockFile(fd):
                        os.write(fd, str(pid).encode())
                        unlockFile(fd)
                        os.close(fd)
                        return True
        except OSError as e:
                if e.errno == errno.EEXIST:
                        #read pid from existing file
                        try:
                                fd=os.open(pidFile, os.O_RDWR|os.O_EXCL|os.O_NONBLOCK)
                                if lockFile(fd):
                                        os.lseek(fd,0,0)
                                        opid=os.read(fd,MAXPIDBYTE)

                                        #Check if opid is currently running process
                                        #otherwise no other instance is running, write current pid
                                        #to the file and return True
                                        try:
                                                if opid.strip() == "":
                                                        opid = MAXPID + 1
                                                else:
                                                        opid = int(opid.decode())
                                                                
                                                os.kill( opid,0)
                                        except OSError as e:
                                                if e.errno == errno.ESRCH:
                                                        os.ftruncate(fd,0)
                                                        os.lseek(fd,0,0)
                                                        spid=str(pid)
                                                        os.write(fd, spid.encode())
                                                        unlockFile(fd)
                                                        os.close(fd)
                                                        return True
                                                else:
                                                        raise OSError(e)

                                        
                                
                                #Else another flashTest process is running or has locked pidFile
                        except OSError as e:
                                #file has been opened by another process
                                if e.errno == errno.EBUSY:
                                        return forceRun
                                elif e.errno == errno.EACCES:
                                        return forceRun
                                else:
                                        raise OSError(e)

                        finally:
                                try: 
                                        os.close(fd)
                                except:
                                        pass

        return forceRun

def unregisterPid(pidFile):
                pid=os.getpid()
                try:
                        #Open file, obtain lock and remove if this flashTest owns it
                        #This will block. Not sure O_EXCL is not ignored
                        fd= os.open(pidFile, os.O_RDONLY|os.O_EXCL) 
                        if lockFile(fd, False):
                                apid = os.read(fd,MAXPIDBYTE)
                                if str(pid) == apid.decode().strip():
                                        unlockFile(fd)
                                        os.close(fd)
                                        os.remove(pidFile)
                                else:
                                        unlockFile(fd)
                                        os.close(fd)
                except Exception as e:
                                pass



#!/usr/bin/env python3

import sys, os, socket
sys.path.insert(0, "lib")
import flashTestParser as parser

# Creates a file "test.info" at the top level of the FlashTest
# directory whose contents are sufficient to run a single UnitTest
# problem. Doing this necessitates knowing the name of the user's
# platform, which is deduced by socket.gethostbyaddr().
# The script will not overwrite an existing "test.info"
pathToFlashTest = os.path.dirname(os.path.abspath(sys.argv[0]))
pathToConfig    = os.path.join(pathToFlashTest, "config")
configDict = parser.parseFile(pathToConfig)
if configDict.has_key("site"):
  site = configDict["site"]
else:
  FQHostname = socket.gethostbyaddr(socket.gethostname())[0]
  site       = FQHostname.split(".",1)[0]

pathToInfoFile  = os.path.join(pathToFlashTest, "test.info")

template = (
"""
<%s>
  <UnitTest>
    setupName: unitTest/Grid/UG
    setupOptions: -1d -auto +noio
    numProcs: 1
    parfiles: <defaultParfile>
  </UnitTest>
</%s>
""")

if os.path.exists(pathToInfoFile):
  print ("This script is only meant to create a brand-new \"test.info\" file,\n" +
         "but the file \"%s\" already exists. Exiting." % pathToInfoFile)
  sys.exit(1)

# else
text = (template % (site, site)).strip()
open(pathToInfoFile, "w").write(text)
print ("Created \"%s\"" % pathToInfoFile)
